@@ -1,6 +1,6 @@
 /* request.cc
  * This file belongs to Worker, a file manager for UN*X/X11.
- * Copyright (C) 2001-2023 Ralf Hoffmann.
+ * Copyright (C) 2001-2025 Ralf Hoffmann.
  * You can contact me at: ralf@boomerangsworld.de
  *   or http://www.boomerangsworld.de/worker
  *
@@ -229,6 +229,269 @@ int Requester::request(const char *title,const char *text,const char *buttons, c
   return endmode;
 }
 
+int Requester::request( const char *title,
+                        const char *text,
+                        const std::vector< std::string > &button_rows,
+                        const AWindow *twin,
+                        request_flags_t flags)
+{
+    std::vector< std::vector< std::pair< std::string, int > > > buttons;
+    int max_buttons_per_row = 1;
+
+    for ( const auto &button_row_string : button_rows ) {
+        char **buttonss;
+        int nbuttons = createLines( button_row_string.c_str(),
+                                    &buttonss );
+
+        if ( nbuttons > max_buttons_per_row ) {
+            max_buttons_per_row = nbuttons;
+        }
+
+        std::vector< std::pair< std::string, int > > button_row;
+        for ( int i = 0; i < nbuttons; i++ ) {
+            button_row.push_back( { buttonss[i], -1 } );
+            _freesafe( buttonss[i] );
+        }
+        _freesafe( buttonss );
+
+        buttons.push_back( button_row );
+    }
+
+    return request( title, text, buttons, twin, flags );
+}
+
+static std::pair< int, int > find_button( const std::vector< std::vector< std::pair< Button *, int > > > &buttons,
+                                          Button *b )
+{
+    std::pair< int, int > res = { -1, -1 };
+
+    for ( size_t row = 0; row < buttons.size(); row++ ) {
+        for ( size_t col = 0; col < buttons[row].size(); col++ ) {
+            if ( buttons[row][col].first == b ) {
+                res = { row, col };
+                break;
+            }
+        }
+    }
+
+    return res;
+}
+
+static int get_end_mode( const std::vector< std::vector< std::pair< Button *, int > > > &buttons,
+                                           size_t col )
+{
+    for ( const auto &row : buttons ) {
+        if ( col >= row.size() ) continue;
+        if ( row.at( col ).first == NULL ) continue;
+        return row.at( col ).second;
+    }
+    return -1;
+}
+
+int Requester::request( const char *title,
+                        const char *text,
+                        const std::vector< std::vector< std::pair< std::string, int > > > &button_rows,
+                        const AWindow *twin,
+                        request_flags_t flags)
+{
+    std::vector< std::vector< std::pair< Button *, int > > > buttons;
+    int max_buttons_per_row = 1;
+
+    buttons.resize( button_rows.size() );
+
+    for ( size_t row = 0; row < button_rows.size(); row++ ) {
+        size_t nbuttons = button_rows[row].size();
+        if ( (int)nbuttons > max_buttons_per_row ) {
+            max_buttons_per_row = nbuttons;
+        }
+    }
+
+    for ( size_t row = 0; row < button_rows.size(); row++ ) {
+        buttons[row].resize( max_buttons_per_row );
+    }
+  
+    if ( twin == NULL && transientawin == NULL ) {
+        twin = aguix->getTransientWindow();
+    }
+  
+    AWindow *win = new AWindow( aguix, 10, 10, 10, 10, title, AWindow::AWINDOW_DIALOG );
+    win->create();
+    if ( twin != NULL ) {
+        win->setTransientForAWindow( twin );
+    } else if ( transientawin != NULL ) {
+        win->setTransientForAWindow( transientawin );
+    }
+
+    AContainer *ac1 = win->setContainer( new AContainer( win, 1, 2 ), true );
+    ac1->setMinSpace( 0 );
+    ac1->setMaxSpace( 0 );
+    ac1->setBorderWidth( 5 );
+
+    AContainerBB *ac1_texts = dynamic_cast<AContainerBB*>( ac1->add( new AContainerBB( win, 1, 1 ), 0, 0 ) );
+    ac1_texts->setBorderWidth( 5 );
+    win->addMultiLineText( text, *ac1_texts, 0, 0, NULL, NULL );
+
+    AContainer *ac1_buttons = NULL;
+
+    if ( max_buttons_per_row == 1 ) {
+        ac1_buttons = ac1->add( new AContainer( win, 3, buttons.size() ), 0, 1 );
+        ac1_buttons->setMinSpace( 0 );
+        ac1_buttons->setMaxSpace( 0 );
+    } else {
+        ac1_buttons = ac1->add( new AContainer( win, max_buttons_per_row, buttons.size() ), 0, 1 );
+        ac1_buttons->setMinSpace( 5 );
+        ac1_buttons->setMaxSpace( -1 );
+    }
+    ac1_buttons->setBorderWidth( 0 );
+
+    size_t row_nr = 0;
+    bool focus_set = false;
+    for ( auto &row : buttons ) {
+        if ( max_buttons_per_row == 1 ) {
+            if ( ! button_rows.at( row_nr ).at( 0 ).first.empty() ) {
+                row.at( 0 ).first = ac1_buttons->addWidget( new Button( aguix, 0, 0, button_rows.at( row_nr ).at( 0 ).first.c_str(), 0 ),
+                                                            1, row_nr, AContainer::CO_FIX );
+                row.at( 0 ).first->setAcceptFocus( true );
+                if ( ! focus_set ) {
+                    row.at( 0 ).first->takeFocus();
+                    focus_set = true;
+                }
+
+                row.at( 0 ).second = button_rows.at( row_nr ).at( 0 ).second;
+                if ( row.at( 0 ).second < 0 ) row.at( 0 ).second = row_nr;
+            }
+        } else {
+            size_t col_nr = 0;
+            for ( auto &col : row ) {
+                if ( col_nr >= button_rows.at( row_nr ).size() ) continue;
+
+                if ( ! button_rows.at( row_nr ).at( col_nr ).first.empty() ) {
+                    col.first = ac1_buttons->addWidget( new Button( aguix, 0, 0, button_rows.at( row_nr ).at( col_nr ).first.c_str(), row_nr * max_buttons_per_row + col_nr ),
+                                                        col_nr, row_nr, AContainer::CO_FIX );
+                    col.first->setAcceptFocus( true );
+                    if ( ! focus_set ) {
+                        col.first->takeFocus();
+                        focus_set = true;
+                    }
+
+                    col.second = button_rows.at( row_nr ).at( col_nr ).second;
+                    if ( col.second < 0 ) col.second = row_nr * max_buttons_per_row + col_nr;
+                }
+
+                col_nr++;
+            }
+        }
+
+        row_nr++;
+    }
+
+    win->setDoTabCycling( true );
+    win->contMaximize( true, true );
+    win->useStippleBackground();
+    win->centerScreen();
+    win->show();
+    AGMessage *msg;
+    int endmode=-1;
+    for(;endmode==-1;) {
+        msg=aguix->WaitMessage(win);
+        if(msg!=NULL) {
+            switch(msg->type) {
+                case AG_CLOSEWINDOW:
+                    if ( msg->closewindow.window == win->getWindow() ) {
+                        if ( ( flags & REQUEST_CANCELWITHLEFT ) != 0 ) {
+                            endmode = get_end_mode( buttons, 0 );
+                        } else {
+                            endmode = get_end_mode( buttons, max_buttons_per_row - 1 );
+                        }
+                    }
+                    break;
+                case AG_BUTTONCLICKED: {
+                    auto res = find_button( buttons, msg->button.button );
+                    if ( res.first >= 0 ) {
+                        endmode = buttons.at( res.first ).at( res.second ).second;
+                        break;
+                    }
+                    break;
+                }
+                case AG_KEYPRESSED:
+                    if ( win->isParent( msg->key.window, false ) == true ) {
+                        switch(msg->key.key) {
+                            case XK_KP_Enter:
+                            case XK_Return: {
+#ifdef ACCEPT_ENTER_FOR_CHOOSE
+                                for ( size_t r = 0; r < buttons.size(); r++ ) {
+                                    for ( size_t c = 0; c < (size_t)max_buttons_per_row; r++ ) {
+                                        if ( c >= buttons[r].size() ) continue;
+                                        if ( buttons[r][c].first == nullptr ) continue;
+                                        if ( buttons[r][c].first->getHasFocus() == true ) {
+                                            endmode = buttons[r][c].second;
+                                            break;
+                                        }
+                                    }
+                                }
+#else
+                                if ( buttons.size() > 0 &&
+                                     buttons[0].size() > 0 &&
+                                     buttons[0][0]->hasFocus() == true ) {
+                                    endmode = buttons[0][0].second;
+                                }
+#endif
+                            }
+                                break;
+                            case XK_Escape:
+                                if ( ( flags & REQUEST_CANCELWITHLEFT ) != 0 ) {
+                                    endmode = get_end_mode( buttons, 0 );
+                                } else {
+                                    endmode = get_end_mode( buttons, max_buttons_per_row - 1 );
+                                }
+                                break;
+                            case XK_F1:
+                                endmode = get_end_mode( buttons, 0 );
+                                break;
+                            case XK_F2:
+                                endmode = get_end_mode( buttons, 1 );
+                                break;
+                            case XK_F3:
+                                endmode = get_end_mode( buttons, 2 );
+                                break;
+                            case XK_F4:
+                                endmode = get_end_mode( buttons, 3 );
+                                break;
+                            case XK_F5:
+                                endmode = get_end_mode( buttons, 4 );
+                                break;
+                            case XK_F6:
+                                endmode = get_end_mode( buttons, 5 );
+                                break;
+                            case XK_F7:
+                                endmode = get_end_mode( buttons, 6 );
+                                break;
+                            case XK_F8:
+                                endmode = get_end_mode( buttons, 7 );
+                                break;
+                            case XK_F9:
+                                endmode = get_end_mode( buttons, 8 );
+                                break;
+                            case XK_F10:
+                                endmode = get_end_mode( buttons, 9 );
+                                break;
+                            case XK_Left:
+                                win->prevFocus();
+                                break;
+                            case XK_Right:
+                                win->nextFocus();
+                                break;
+                        }
+                    }
+                    break;
+            }
+            aguix->ReplyMessage(msg);
+        }
+    }
+    delete win;
+    return endmode;
+}
+
 int Requester::string_request( const char *title, const char *text, const char *default_str, const char *buttons, char **return_str )
 {
   return string_request( title, text, default_str, buttons, return_str, REQUEST_NONE );
@@ -240,14 +503,25 @@ int Requester::string_request( const char *title, const char *text, const char *
 }
 
 int Requester::string_request( const char *title, const char *text, const char *default_str, const char *buttons,char **return_str, const AWindow *twin, request_flags_t flags)
+{
+    request_options options;
+
+    return string_request( title, text, default_str,
+                           buttons, return_str, twin, flags, options );
+}
+
+int Requester::string_request( const char *title, const char *text, const char *default_str, const char *buttons,char **return_str, const AWindow *twin, request_flags_t flags,
+                               struct request_options &options )
 {
     return string_request( title, text, default_str,
                            buttons, return_str, twin, flags,
+                           options,
                            "", nullptr );
 }
 
 int Requester::string_request( const char *title, const char *text, const char *default_str, const char *buttons,
                                char **return_str, const AWindow *twin, request_flags_t flags,
+                               struct request_options &options,
                                const std::string &help_text,
                                const std::function< std::string ( const std::string &current_input_before_cursor,
                                                                   const std::string &current_input_after_cursor ) > help_callback )
@@ -341,6 +615,12 @@ int Requester::string_request( const char *title, const char *text, const char *
     }
   }
 
+  if ( options.select_characters < 0 ) {
+      sg->selectAll();
+  } else if ( options.select_characters > 0 ) {
+      sg->setSelection( 0, options.select_characters - 1, false );
+  }
+
   sg->takeFocus();
   win->useStippleBackground();
   win->setDoTabCycling( true );
